from rune.context import Context
from rune.exception.notfounderror import NotFoundError
from rune.models.result import Failure, Result, Success
from rune.models.crypto.secret import Secret

from typing import List

def list_secrets(user: str) -> Result[List[Secret]]:
    """
    Retrieves all secret entries with the configured storage manager.
    Returns None if it there is an error getting the secrets.
    """
    storage = Context.get().storage_manager
    try:
        return Success(storage.get_all_secrets(user))
    except NotFoundError as err:
        return Failure(err.message)




from rich.prompt import Prompt

from rune.internal.move import move_secret
from rune.utils.input import sanitize_name
from rune.utils import display

def handle_move_command(user: str, _original_name: str | None, _new_name: str | None):
    original_name = sanitize_name(_original_name or Prompt.ask("The name of the secret you want to move"))
    new_name = sanitize_name(_new_name or Prompt.ask("The new name of the secret"))

    result = move_secret(user, original_name, new_name)

    if result.is_success():
        display.success_panel(f"Secret '[bold cyan]{original_name}[/]' was moved to '[bold cyan]{new_name}[/]'.")
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")





from abc import ABC, abstractmethod
from typing import List, Optional

from rune.models.crypto.secret import Secret

class StorageManager(ABC):
    @abstractmethod
    def store_secret(self, secret: Secret) -> bool:
        """
        Stores the provided secret.

        Returns True if storage is successful, False otherwise.
        Raises NotFoundError if it fails to find a secrets file.
        """
        raise NotImplementedError()

    @abstractmethod
    def retreive_secret(self, user: str, name: str)-> Optional[Secret]:
        """
        Retreives the provided ciphertext under the provided secret name under the provided user.

        Raises NotFoundError if it fails to find a secrets file.
        """
        raise NotImplementedError()

    @abstractmethod
    def move_secret(self, user: str, original_name: str, new_name: str) -> bool:
        """
        Moves a secret from one name to another. Does not change encryption.
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_secret(self, user: str, name: str, hard: bool) -> bool:
        """
        Deletes the entry with the provided name for the provided user.

        Returns True if successful, False if it fails.
        Raises NotFoundError if it fails to find a secrets file.
        """
        raise NotImplementedError()

    @abstractmethod
    def restore_secret(self, user: str, name: str) -> bool:
        """
        Restores a soft deleted secret (makes it not soft deleted)
        """
        raise NotImplementedError()

    @abstractmethod
    def get_all_secrets(self, user: str) -> List[Secret]:
        """
        Retrieves all entry names for a specific user.

        Raises NotFoundError if it fails to retreive entries.
        """
        raise NotImplementedError()





from typing import Dict, Optional, List
from rune.models.crypto.secret import Secret
from rune.storage.secrets.base import StorageManager
import json
from pathlib import Path

class LocalJsonStorageManager(StorageManager):
    def __init__(self, secrets_file_path: Path) -> None:
        self.__secrets_file_path: Path = secrets_file_path
        self._ensure_secrets()

    def _ensure_secrets(self) -> None:
        if not self.__secrets_file_path.exists():
            self.__secrets_file_path.parent.mkdir(parents=True, exist_ok=True)
            with open(self.__secrets_file_path, "x") as f:
                json.dump({}, f, indent=4)

    def store_secret(self, secret: Secret) -> bool:
        """
        Stores the provided ciphertext under the provided secret name.

        Returns True if storage is successful, False otherwise.
        Raises NotFoundError if it fails to find a secrets file.
        """
        secrets = self.stored_secrets_by_full_name(secret.user)
        secrets[secret.full_name] = secret

        return self.store_secrets(secrets)

    def retreive_secret(self, user: str, name: str) -> Optional[Secret]:
        """
        Retreives the provided ciphertext under the provided secret name.

        Raises NotFoundError if it fails to find a secrets file.
        """
        secrets = self.stored_secrets_by_full_name(user)
        return secrets.get(name)

    def move_secret(self, user: str, original_name: str, new_name: str) -> bool:
        """
        Moves a secret from one name to another. Does not change encryption.
        """
        original = self.retreive_secret(user, original_name)
        if not original:
            return False

        new_secret = original.update(full_name=new_name)
        updated = {k: v for k, v in self.stored_secrets_by_full_name(user).items() if not k == original_name}
        updated[new_name] = new_secret

        return self.store_secrets(updated)

    def delete_secret(self, user: str, name: str, hard: bool) -> bool:
        """
        Deletes the entry with the provided name.

        Returns True if successful, False if it fails.
        Raises NotFoundError if it fails to find a secrets file.
        """
        secrets = self.stored_secrets_by_full_name(user)

        if not name in secrets:
            return False
        
        if hard:
            secrets = {n: s for n, s in secrets.items() if not n == name}
        else:
            secrets[name] = secrets[name].soft_delete()

        return self.store_secrets(secrets)

    def restore_secret(self, user: str, name: str) -> bool:
        """
        Restores a soft deleted secret (makes it not soft deleted)
        """
        secrets = self.stored_secrets_by_full_name(user)

        to_store = secrets.get(name)

        if to_store and to_store.deleted:
            return self.store_secret(to_store.restore())
        
        return False


    def get_all_secrets(self, user: str) -> List[Secret]:
        """
        Retrieves all entry names.
        """
        with open(self.__secrets_file_path, "r") as f:
            d = json.load(f)
            all_secrets = [ Secret.from_dict(v) for _, v in d.items() ]
            return [s for s in all_secrets if s.user == user]


    def store_secrets(self, secrets: Dict[str, Secret]) -> bool:
        try:
            with open(self.__secrets_file_path, "w") as f:
                to_dump = {s.id: s.to_dict() for s in secrets.values()}
                json.dump(to_dump, f, indent=4)
                return True
        except:
            return False


    def stored_secrets_by_full_name(self, user: str) -> Dict[str, Secret]:
        secrets = self.get_all_secrets(user)
        return {s.full_name: s for s in secrets}
        





from typing import Annotated, Optional
import typer
from typer import Typer

from rune.commands.addcmd import handle_add_cmd
from rune.commands.getcmd import handle_get_command
from rune.commands.movecmd import handle_move_command
from rune.commands.updatecmd import handle_update_command
from rune.commands.deletecmd import handle_delete_command, handle_restore_cmd
from rune.commands.listcmd import handle_ls_command
from rune.utils.input import ensure_active_user

NAME_HELP = (
    "The name of the secret. Supports namespaces (e.g., `db/prod/my-db`). "
    "If omitted, you will be prompted."
)

FIELDS_HELP = (
    "Comma-separated fields to store. You can provide `key=value` pairs "
    "for known values. Fields without a value will be queried securely.\n"
    "If omitted, will store a single-field secret where the field has the same name as the secret.\n"
    "Example: -f host=localhost,port=9999,user,password"
)

KEY_HELP = (
    "Encryption key to use. If omitted, the active session key will be used (if available).\n"
    "If no active session key is found, it will be queried securely."
)

KEY_HELP_UPDATE = (
    "Encryption key to use. Should be the same as the one used to originally encrypt the secret\n"
    "If omitted, the active session key will be used (if available).\n"
    "If no active session key is found, it will be queried securely."
)

def setup(app: Typer):

    @app.command()
    def add(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _fields: Annotated[Optional[str], typer.Option("--fields", "-f", help=FIELDS_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP)] = None
    ):
        """
        Add a new secret to the vault.

        Supports namespaced secrets. Fields without explicit values
        will be prompted securely.
        """
        active_user = ensure_active_user()
        handle_add_cmd(active_user, _fields, _name, _key)

    @app.command()
    def delete(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _hard: Annotated[bool, typer.Option("--hard", help="Hard delete the secret. Requires encryption key.")] = False,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help="Key used to encrypt the secret. Required if `--hard`.")] = None,
    ):
        """
        Delete a secret from the vault.

        Hard delete requires original encryption key.
        """
        active_user = ensure_active_user()
        handle_delete_command(active_user, _hard, _name, _key)

    @app.command()
    def restore(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None
    ):
        """
        Restore a soft deleted secret.
        """
        active_user = ensure_active_user()
        handle_restore_cmd(active_user, _name)


    @app.command()
    def update(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _fields: Annotated[Optional[str], typer.Option("--fields", "-f", help=FIELDS_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP_UPDATE)] = None
    ):
        """
        Update an existing secret in the vault.

        Fields without explicit values will be prompted securely.
        """
        active_user = ensure_active_user()
        handle_update_command(active_user, _fields, _name, _key)

    @app.command()
    def get(
        _name: Annotated[Optional[str], typer.Argument(help=NAME_HELP)] = None,
        _key: Annotated[Optional[str], typer.Option("--key", "-k", help=KEY_HELP)] = None,
        interactive: Annotated[bool, typer.Option(
            "--interactive", "-i",
            help="Shortcut for `rune ls -i`. Name and key are ignored when using interactive."
        )] = False,
        show: Annotated[bool, typer.Option("--show","-s",help="Show secret values in the terminal instead of hiding them.")]=False
    ):
        """
        Retrieve a secret from the vault.

        Copies the selected field to clipboard by default.
        """
        active_user = ensure_active_user()
        if interactive:
            handle_ls_command(active_user, namespace=None, interactive=True, show=show, show_deleted=False)
        else:
            handle_get_command(active_user, _name, _key, show)

    OG_NAME_HELP = "Full name of secret to move"
    DEST_NAME_HELP = "Destination name for the secret"
    @app.command(name="move")
    def move(
        _original_name: Annotated[Optional[str], typer.Argument(help=OG_NAME_HELP)] = None,
        _new_name: Annotated[Optional[str], typer.Argument(help=DEST_NAME_HELP)] = None
    ):
        """
        Move a secret from one name to another.
        """
        active_user = ensure_active_user()
        handle_move_command(active_user, _original_name, _new_name)

    @app.command(name="ls")
    def list_entries(
        namespace: Annotated[Optional[str], typer.Argument(help="Filter secrets by namespace")] = None,
        interactive: Annotated[bool, typer.Option("--interactive", "-i", help="Interactively select and retrieve secrets from the list.")] = False,
        show: Annotated[bool, typer.Option("--show","-s", help="Show secret values in the terminal. Only used with --interactive.")]=False,
        show_deleted: Annotated[bool, typer.Option("--show-deleted", help="Show soft deleted secrets.")] = False
    ):
        """
        List all secrets in the vault for the logged in user, organized by namespace.

        Single-child namespaces are collapsed for cleaner display.
        Use `--namespace` to filter results.
        """
        active_user = ensure_active_user()
        handle_ls_command(active_user, namespace, interactive, show, show_deleted)




from abc import ABC, abstractmethod
from typing import Dict, Self
from dataclasses import dataclass
import os

from platformdirs import PlatformDirs

class StorageSettings(ABC):
    FILE_BASED: str = "local"

    @property
    def mode(self) -> str: raise NotImplementedError()

    @abstractmethod
    def to_dict(self) -> Dict: raise NotImplementedError()

    @classmethod
    def from_dict(cls, d: Dict) -> "StorageSettings":
        match d["mode"]:
            case cls.FILE_BASED:
                return FileBasedStorageSettings.from_dict(d)
            case _:
                raise ValueError()

    @classmethod
    def default(cls) -> "StorageSettings":
        return FileBasedStorageSettings.default()

@dataclass
class FileBasedStorageSettings(StorageSettings):
    file: str

    @property
    def mode(self) -> str: return StorageSettings.FILE_BASED

    def to_dict(self) -> Dict:
        return {
            "mode": self.mode,
            "file": self.file
        }
    
    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(d["file"])

    @classmethod
    def default(cls) -> Self:
        dirs = PlatformDirs("rune", None)
        return cls(
            file = os.path.join(dirs.user_config_dir, "secrets.json")
        )






from rich.tree import Tree

def dict_to_rich_tree(data: dict, label: str = "root") -> Tree:
    tree = Tree(label)

    def _add(parent: Tree, node: dict):
        for key, value in node.items():
            if isinstance(value, dict):
                branch = parent.add(f"[bold cyan]{key}[/]")
                _add(branch, value)
            elif isinstance(value, list):
                list_branch = parent.add(f"[bold green]{key}[/]")
                for item in value:
                    list_branch.add(str(item))
            else:
                parent.add(f"{key}: {value}")

    _add(tree, data)
    return tree



from rune.context import Context
from rune.exception.notfounderror import NotFoundError
from rune.exception.wrongencryption import WrongEncryptionMode
from rune.exception.wrongkey import WrongKeyUsed
from rune.models.result import Result, Success, Failure
from rune.crypto import factory as EncrypterFactory

from typing import Dict

def get_secret(user: str, full_name: str, key: str) -> Result[Dict[str, str]]:
    """
    Retreives the encrypted secret via the configured storage manager.
    Decrypts the ciphertext with the provided key.

    Returns the decrypted secret, if it exists.
    Returns None if not successful.
    """
    storage = Context.get().storage_manager

    try:
        secret = storage.retreive_secret(user, full_name)
        if secret is not None:
            try:
                decrypted_fields = {}
                for field_name, field in secret.fields.items():
                    encrypter = EncrypterFactory.get_encrypter_by_algorithm(field.algorithm)
                    decrypted_fields[field_name] = encrypter.decrypt(field, key)
            except WrongEncryptionMode as err:
                return Failure(err.message)
            except WrongKeyUsed as err:
                return Failure(err.message)

            return Success(decrypted_fields)
        else:
            return Failure(f"Secret '{full_name}' does not exist.")

    except NotFoundError as err:
        return Failure(err.message)





from typing import Self
from rune.crypto.base import Encrypter
from rune.crypto.factory import get_encrypter
from rune.models.settings.settings import Settings
from rune.storage.secrets.base import StorageManager
from rune.storage.secrets.factory import get_storage_manager
from rune.storage.settings.settingsstorage import SettingsStorageManager

class Context:
    _context: Self | None = None

    def __init__(
        self,
        settings_manager: SettingsStorageManager | None,
        encrypter: Encrypter | None,
        storage_manager: StorageManager | None

    ) -> None:
        self.settings_manager: SettingsStorageManager = settings_manager or SettingsStorageManager()
        self.settings: Settings = self.settings_manager.load_settings()
        self.configured_encrypter: Encrypter = encrypter or get_encrypter(self.settings)
        self.storage_manager: StorageManager = storage_manager or get_storage_manager(self.settings)

    @classmethod
    def get(cls) -> Self:
        if cls._context is not None:
            return cls._context
        raise RuntimeError("Context is not set. Call `Context.build()` first.")

    @classmethod
    def build(
        cls,
        settings_manager: SettingsStorageManager | None = None,
        encrypter: Encrypter | None = None,
        storage_manager: StorageManager | None = None,
    ) -> Self:
        if not cls._context:
            cls._context = cls(settings_manager, encrypter, storage_manager)
            return cls._context
        raise RuntimeError("Context is already built")

    @classmethod
    def update(
        cls,
        settings_manager: SettingsStorageManager | None = None,
        encrypter: Encrypter | None = None,
        storage_manager: StorageManager | None = None,
    ) -> Self:
        if cls._context:
            cls._context = cls(settings_manager, encrypter, storage_manager)
            return cls._context
        raise RuntimeError("Context is not set. Call `Context.build()` first.")


    @classmethod
    def reset(cls) -> None:
        cls._context = None





from rune.context import Context
from rune.exception.notfounderror import NotFoundError
from rune.models.result import Failure, Result, Success
from rune.models.crypto.secret import Secret

from typing import Dict


def add_secret(user: str, full_name: str, fields: Dict[str, str], key: str) -> Result[None]:
    """
    Encrypts a secret with the configured encrypter.
    Stores the encrypted secret with the configured storage manager.

    Returns the result.
    """
    encrypter = Context.get().configured_encrypter
    storage = Context.get().storage_manager

    encrypted_fields = {name: encrypter.encrypt(secret, key) for name, secret in fields.items()}

    model = Secret(
        full_name = full_name,
        user = user,
        algorithm = encrypter._encryption_algorithm,
        fields = encrypted_fields
    )

    fqn = model.full_name

    try:
        if storage.retreive_secret(user, fqn) is not None:
            return Failure(f"Secret '{fqn}' already exists. You can update it with `rune update -n {fqn}`")

        return Success() if storage.store_secret(model) else Failure(f"Storage manager could not store the secret {fqn}")


    except NotFoundError as err:
        return Failure(err.message)




from rune.internal.get import get_secret
from rune.utils.input import input_key, input_name, sanitize_name
from rune.utils import display

from rich.console import Console
from rich.panel import Panel
from rich.prompt import Prompt
import pyperclip

console = Console()

def handle_get_command(user: str, _name: str | None = None, _key: str | None = None, show: bool = False):
    name = sanitize_name(_name or input_name())
    key = (_key or input_key())

    result = get_secret(user, name, key)

    v = result.value()

    if not (result.is_success() and v):
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")
        return


    console.print(Panel.fit(
        f"Stored fields for secret '[bold cyan]{name}[/]':" + "\n" + \
        "\n".join([f"[bold cyan][{i}][/] {k}" + ("" if not show else f" - {s}") for i, (k, s) in enumerate(v.items(), 1)])
    ))

    keys = list(v.keys())

    while True:
        choice = Prompt.ask(
            "[cyan]Select field to copy[/] (q to cancel)",
        )
        if choice.lower() == "q":
            break
        try:
            index = int(choice) - 1
            if 0 <= index < len(keys):
                selected_key = keys[index]
                pyperclip.copy(v[selected_key])
                display.success_panel(f"[bold green]✓ Copied[/] [yellow]{selected_key}[/] to clipboard", title="Copied")
        except:
            pass





from abc import ABC, abstractmethod

from rune.models.crypto.secretfield import SecretField


class Encrypter(ABC):
    def __init__(self) -> None:
        self._encryption_algorithm: str

    @abstractmethod
    def encrypt(self, secret: str, key: str, **kwargs) -> SecretField:
        """
        Encrypts the provided secret with the provided key.
        """
        raise NotImplementedError()

    @abstractmethod
    def decrypt(self, secret: SecretField, key: str, **kwargs) -> str:
        """
        Decrypts the provided secret with the provided key
        """
        raise NotImplementedError()



from rich.console import Console
from rich.panel import Panel

from rune.internal.add import add_secret
from rune.utils.input import input_key, input_name, sanitize_name, get_fields_dict
from rune.utils import display

console = Console()

def handle_add_cmd(user: str, _fields: str | None, _name: str | None, _key: str | None):
    name = sanitize_name(_name or input_name())
    fields = get_fields_dict(_fields, name)
    key = _key or input_key()

    result = add_secret(user, name, fields, key)

    if result.is_success():
        display.success_panel(f"[bold green]✓ Stored new secret[/] [cyan]{name}[/]")
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")





from rune.context import Context
from rune.utils import display


def handle_login_command(username: str) -> None:
    Context.get().settings.update(active_user=username)
    display.success_panel(f"Logged in as [bold]'{username}'[/]")

def handle_logout_command() -> None:
    active_user = Context.get().settings.active_user

    if not active_user:
        display.failed_panel("Not logged in")
        return

    Context.get().settings.reset(active_user=True)

    display.success_panel("Successfully logged out")






import json
from typing import Dict, Optional
from rune.models.settings.settings import Settings

from platformdirs import PlatformDirs
import os
from pathlib import Path


class SettingsStorageManager:

    def __init__(self, config_dir: Path | None = None, settings_file: Path | None = None, profiles_file: Path | None = None) -> None:
        self.config_dir: Path = config_dir or Path(PlatformDirs("rune", None).user_config_dir)
        self.settings_file: Path = settings_file or Path(os.path.join(self.config_dir, "settings.json"))
        self.profiles_file: Path = profiles_file or Path(os.path.join(self.config_dir, "profiles.json"))
        self._ensure_settings()

    def _ensure_settings(self) -> None:
        if not self.settings_file.exists():
            self.settings_file.parent.mkdir(parents=True, exist_ok=True)
            with open(self.settings_file, "x") as f:
                json.dump(Settings.default().to_dict(), f, indent=4)

    def _ensure_profiles(self) -> None:
        if not self.profiles_file.exists():
            self.profiles_file.parent.mkdir(parents=True, exist_ok=True)
            with open(self.profiles_file, "x") as f:
                json.dump({}, f, indent=4)

    def load_settings(self) -> Settings:
        with open(self.settings_file, "r") as f:
            d = json.load(f)
            return Settings.from_dict(d)
    
    def save_settings(self, settings: Settings) -> None:
        with open(self.settings_file, "w") as f:
            json.dump(settings.to_dict(), f, indent=4)

    def get_profiles(self) -> Dict[str, Settings]:
        self._ensure_profiles()
        with open(self.profiles_file, "r") as f:
            profile_d = json.load(f)
            return {name: Settings.from_dict(settings) for name, settings in profile_d.items()}

    def get_profile(self, name: str) -> Optional[Settings]:
        profiles = self.get_profiles()
        profile: Settings | None = profiles.get(name)
        return profile if profile else None

    def delete_profile(self, name: str) -> None:
        profiles = self.get_profiles()
        removed = {k: v for k, v in profiles.items() if not name == k}
        self._store_profiles(removed)

    def save_profile(self, settings: Settings, name: str) -> None:
        profiles = self.get_profiles()
        profiles[name] = settings
        self._store_profiles(profiles)

    def _store_profiles(self, profiles: Dict[str, Settings]) -> None:
        self._ensure_profiles()
        with open(self.profiles_file, "w") as f:
            profile_dict = {name: profile.to_dict() for name, profile in profiles.items()}
            json.dump(profile_dict, f, indent=4)








from rune.session.base import SessionManager
from typing import Optional


class MockSessionManager(SessionManager):
    def start_session(self, default_password: Optional[str] = None, ttl_seconds: int = -1) -> None:
        """
        Starts a session.
        Session exists for the provided ttl. (-1 means it will not close)
        """
        pass

    def end_session(self) -> None:
        """
        Ends the current session.

        raises NoSessionError if the session does not exist.
        """
        pass

    def is_session_in_progress(self) -> bool:
        """
        Returns true if there is an ongoing session

        False otherwise.
        """
        return True

    def set_default_key(self, key: str) -> None:
        """
        Sets the default key for this session.

        raises NoSessionError if the session does not exist.
        """
        pass

    def get_default_key(self) -> Optional[str]:
        """
        Retrieves the default key for this session.
        Returns None if the key is not set.

        Raises NoSessionError if the session does not exist.
        """
        return "default-key"

    def is_default_key_defined(self) -> bool:
        """
        Raises NoSessionError if the session does not exist
        """
        return True




from abc import ABC, abstractmethod
from typing import Dict, Self
from dataclasses import dataclass

class EncryptionSettings(ABC):
    AES_GCM: str = "aesgcm"

    @property
    def mode(self) -> str: raise NotImplementedError()

    @abstractmethod
    def to_dict(self) -> Dict: raise NotImplementedError()

    @classmethod
    def from_dict(cls, d: Dict) -> "EncryptionSettings":
        match d["mode"]:
            case cls.AES_GCM:
                return AES_GCMEncryptionSettings.from_dict(d)
            case _:
                raise ValueError(f"Unknown encryption mode '{d["mode"]}'")

    @classmethod
    def from_mode(cls, mode: str) -> "EncryptionSettings":
        match mode:
            case cls.AES_GCM:
                return AES_GCMEncryptionSettings.default()
            case _:
                raise ValueError(f"Unknown encryption mode '{mode}'")

    @classmethod
    def default(cls) -> "EncryptionSettings":
        return AES_GCMEncryptionSettings.default()



@dataclass
class AES_GCMEncryptionSettings(EncryptionSettings):
    @property
    def mode(self) -> str: return "aesgcm"

    def to_dict(self) -> Dict:
        return {
            "mode": self.mode,
        }
    
    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls()

    @classmethod
    def default(cls) -> Self:
        return cls()





from typing import List, Tuple, Dict

from rune.context import Context
from rich.prompt import Prompt

from rune.exception.badinput import BadInputError
from rune.models.crypto.namespace import Namespace

NAME_PROMPT = "Secret name"
KEY_PROMPT = "Encryption key"

def input_name() -> str:
    return Prompt.ask(NAME_PROMPT)

def input_key() -> str:
    return Prompt.ask(KEY_PROMPT, password=True)

def input_default_key() -> str:
    return Prompt.ask("The [bold]default key[/] for this session")

def sanitize_name(full_name: str) -> str:
    if "//" in full_name:
        raise BadInputError(f"Not valid name. Please input a '/'-separated name. [bold]Ex: db/prod/service[/]")
    return full_name.removeprefix("/").removesuffix("/")

def get_last_part_of_name(full_name: str) -> str:
    return full_name.split("/")[-1]

def split_name_and_ns(n_and_ns: str) -> Tuple[str, Namespace]:
    s = n_and_ns.split("/")
    return s[-1], Namespace(s[:-1])

def get_secret_input(name: str) -> str:
    return Prompt.ask(f"Value for field '[bold]{name}[/]'", password=True)

def get_fields_dict(fields: str | None, name: str) -> Dict[str, str]:
    ret = {}
    fields = fields or name.split("/")[-1]
    for field in fields.split(","):
        split = field.split("=")
        if len(split) == 1:
            ret[field] = get_secret_input(field)
        elif len(split) >= 2:
            ret[split[0]] = "".join(split[1:])
    return ret

def get_fqn(name: str, namespace: Namespace) -> str:
    return "/".join(namespace.ns + [name])

def get_active_user() -> str | None:
    return Context.get().settings.active_user

def ensure_active_user() -> str:
    maybe_user = get_active_user()
    if not maybe_user:
        raise BadInputError("[bold red]User not set.[/] Please log in with [bold]`rune login -u <username>`[/]")

    return maybe_user

def get_choice_by_idx[T](prompt: str, elements: List[T]) -> T | None:
    while True:
        choice = Prompt.ask(f"{prompt} (q to quit)")
        if choice.lower() == "q":
            return None
        try:
            idx = int(choice)
            return elements[idx - 1]
        except:
            continue

def require[T](arg: T | None, reason: str = "") -> T:
    if arg is None:
        raise BadInputError(reason)
    return arg





from rune.context import Context
from rune.crypto import factory as EncryptionFactory
from rune.exception.notfounderror import NotFoundError
from rune.exception.wrongkey import WrongKeyUsed
from rune.models.result import Failure, Result, Success

from typing import Dict

def update_secret(user: str, full_name: str, fields: Dict[str, str], key: str) -> Result[None]:
    """
    Encrypts a secret with the configured encrypter.
    Updates the encrypted secret (if it exists) with the configured storage manager.

    Returns the result.
    """
    storage = Context.get().storage_manager

    try:
        original_secret = storage.retreive_secret(user, full_name)
        decrypted_fields = {}
        if original_secret is not None:
            for full_name, field in original_secret.fields.items():
                encrypter = EncryptionFactory.get_encrypter_by_algorithm(field.algorithm)
                decrypted_fields[full_name] = encrypter.decrypt(field, key)
        else:
            return Failure(f"Secret '{full_name}' does not exist. You can create it with `rune add -n {full_name}`.")
    except WrongKeyUsed as err:
        return Failure(f"You have to use the same key to update a secret.")

    encrypter = Context.get().configured_encrypter
    provided_encrypted_fields = {name: encrypter.encrypt(secret, key) for name, secret in fields.items()}

    encrypted_fields = {}
    for full_name, f in original_secret.fields.items():
        encrypted_fields[full_name] = provided_encrypted_fields.get(full_name) or f

    for full_name, f in provided_encrypted_fields.items():
        if full_name not in original_secret.fields:
            encrypted_fields[full_name] = f

    model = original_secret.update(
        algorithm = encrypter._encryption_algorithm,
        fields = encrypted_fields
    )
    try:
        if storage.store_secret(model):
            return Success()
        else:
            return Failure(f"Storage manager could not store the secret {full_name}.")

    except NotFoundError as err:
        return Failure(err.message)




from rich.console import Console
from rich.panel import Panel
import typer
from rune.cli import configmanagement, secretmanagement, usermanagement
from rune.context import Context
from rune.exception.badinput import BadInputError

error_management = "ignore"

def main():
    app = typer.Typer(context_settings={"help_option_names": ["-h", "--help"]})
    config_app = typer.Typer(name="config", help="Manage rune configs. Run `rune config -h` for more help.")

    secretmanagement.setup(app)
    usermanagement.setup(app)
    configmanagement.setup(config_app)

    app.add_typer(config_app)

    console = Console()

    try:
        Context.build()
        app()

    except BadInputError as e:
        console.print(Panel.fit(e.message, title="[red]Bad input[/]"))
    except RuntimeError as e:
        if error_management == "rethrow":
            raise e
        elif error_management == "print":
            print(e)
    finally:
        shutdown()

def shutdown():
    settings = Context.get().settings
    if settings._dirty:
        Context.get().settings_manager.save_settings(Context.get().settings)





from rune.internal.update import update_secret
from rune.utils.input import input_key, input_name, sanitize_name, get_fields_dict
from rune.utils import display

def handle_update_command(user: str, _fields: str | None, _name: str | None, _key: str | None):

    name = sanitize_name(_name or input_name())
    fields = get_fields_dict(_fields, name)
    key = (_key or input_key())
    result = update_secret(user, name, fields, key)

    if result.is_success():
        display.success_panel(f"[bold green]Updated secret[/] [cyan]{name}[/]")
    else:
        display.failed_panel(f"[bold red]Error:[/] {result.failure_reason()}")





from dataclasses import dataclass
from typing import Dict, Optional, Self

from rune.models.settings.encryptionsettings import EncryptionSettings
from rune.models.settings.storagesettings import StorageSettings


@dataclass
class Settings:
    encryption: EncryptionSettings
    storage: StorageSettings
    active_user: Optional[str]

    _dirty: bool = False

    def to_dict(self) -> Dict:
        ret: Dict = {
            "encryption": self.encryption.to_dict(),
            "storage": self.storage.to_dict()
        }
        if self.active_user:
            ret["active_user"] = self.active_user
        return ret

    def update(self, encryption: EncryptionSettings | None = None, storage: StorageSettings | None = None, active_user: str | None = None) -> None:
        self.encryption = encryption or self.encryption
        self.storage = storage or self.storage
        self.active_user = active_user or self.active_user
        self._dirty = True

    def reset(self, encryption: bool = False, storage: bool = False, active_user: bool = False) -> None:
        self._dirty = True
        default = self.default()
        if encryption:  self.encryption = default.encryption
        if storage:     self.storage = default.storage
        if active_user: self.active_user = default.active_user

    def dirty(self) -> Self:
        self._dirty = True
        return self

    @classmethod
    def from_dict(cls, d: Dict) -> Self:
        return cls(
            encryption = EncryptionSettings.from_dict(d["encryption"]),
            storage = StorageSettings.from_dict(d["storage"]),
            active_user = d.get("active_user", None)
        )

    @classmethod
    def default(cls) -> Self:
        return cls(
            encryption = EncryptionSettings.default(),
            storage = StorageSettings.default(),
            active_user = None,
        )





from pathlib import Path
from rich.console import Console
from rich.panel import Panel
from rich.pretty import Pretty
from typer import Argument, Typer, Option
from typing import Annotated, Literal, Optional
import pyperclip
import typer

from rune.context import Context
from rune.models.settings.encryptionsettings import EncryptionSettings
from rune.models.settings.storagesettings import FileBasedStorageSettings
from rune.utils.input import get_choice_by_idx, require
from rune.utils import display

STORAGE_MODE_HELP = (
    "Configure how rune stores encrypted secrets.\n"
    "Currently supported modes:\n"
    "  - local: store secrets in a local file on disk."
)

STORAGE_FILE_HELP = (
    "Path to the secrets file when using local storage.\n"
    "Example: ~/.rune/secrets.json"
)

ENCRYPTION_MODE_HELP = (
    "Configure the encryption algorithm used to encrypt secrets.\n"
    "Changing this does NOT re-encrypt existing secrets."
)

console = Console()


def setup(app: Typer):
    """
    Register configuration-related commands.
    """

    profile_app = Typer(
        name="profile",
        help="Manage configuration profiles (save, switch, list, delete).",
    )

    @profile_app.command(name="list")
    def list_profiles(
        interactive: Annotated[
            bool,
            Option(
                "--interactive",
                "-i",
                help="Interactively select a profile to activate after listing.",
            ),
        ] = False
    ):
        """
        List all configured profiles.
        """
        settings_manager = Context.get().settings_manager

        profiles = list(settings_manager.get_profiles().keys())
        profiles_file = str(settings_manager.profiles_file.absolute())

        console.print(f"[bold]Profiles file:[/] [cyan]{profiles_file}[/]")

        if not profiles:
            console.print(
                Panel.fit(
                    "[yellow]No profiles configured yet.[/]\n\n"
                    "Create one with:\n"
                    "[bold cyan]rune config profile save <profile-name>[/]"
                )
            )
            return

        console.print(
            Panel.fit(
                "\n".join(
                    f"[bold cyan][{idx}][/] {profile}"
                    for idx, profile in enumerate(profiles, 1)
                ),
                title="Available Profiles",
            )
        )

        if not interactive:
            return

        choice = get_choice_by_idx("Select profile to activate", profiles)
        if not choice:
            return

        load_profile(choice)

    @profile_app.command(name="save")
    def save_profile(
        _name: Annotated[
            str, Argument(help="Name under which the current configuration will be saved.")
        ],
        _force: Annotated[
            bool,
            Option(
                "--force",
                "-f",
                help="Overwrite the profile if it already exists.",
            ),
        ] = False,
    ):
        """
        Save the current configuration as a profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        if _name in settings_manager.get_profiles() and not _force:
            console.print(
                Panel.fit(
                    f"Profile '[bold cyan]{_name}[/]' already exists.\n\n"
                    f"Use [bold cyan]--force[/] to overwrite it.",
                    title="[red]Failed[/]",
                )
            )
            return

        settings_manager.save_profile(context.settings, _name)
        display.success_panel(f"Profile '[bold cyan]{_name}[/]' saved successfully.")

    @profile_app.command(name="load")
    def load_profile(
        _name: Annotated[
            str, Argument(help="Name of the profile to activate.")
        ],
    ):
        """
        Activate a saved configuration profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        settings = settings_manager.get_profile(_name)
        if not settings:
            display.failed_panel(f"Profile '[bold cyan]{_name}[/]' does not exist.")
            return

        context.settings = settings.dirty()
        settings_manager.save_profile(settings, _name)

        display.success_panel(f"Switched to profile '[bold cyan]{_name}[/]'.")

    @profile_app.command(name="delete")
    def delete_profile(
        _name: Annotated[
            str, Argument(help="Name of the profile to delete.")
        ],
    ):
        """
        Delete an existing profile.
        """
        context = Context.get()
        settings_manager = context.settings_manager

        profile = settings_manager.get_profile(_name)
        if not profile:
            display.failed_panel(f"Profile '[bold cyan]{_name}[/]' does not exist.")
            return

        if typer.confirm(f"Are you sure you want to delete profile '{_name}'?"):
            settings_manager.delete_profile(_name)
            display.success_panel(f"Profile '[bold cyan]{_name}[/]' deleted.")

        else:
            raise typer.Abort()

    @app.command(name="storage")
    def config_storage(
        _mode: Annotated[
            Optional[Literal["local"]],
            Option("--mode", "-m", help=STORAGE_MODE_HELP),
        ] = None,
        _file: Annotated[
            Optional[str],
            Option("--file", "-f", help=STORAGE_FILE_HELP),
        ] = None,
    ):
        """
        Configure how and where secrets are stored.
        """
        context = Context.get()
        mode: str = _mode or context.settings.storage.mode

        if mode == "local":
            file = require(_file, "A file path is required when using local storage.")
            path = Path(file)
            storage_path = str(path.expanduser().absolute())

            new_settings = FileBasedStorageSettings(storage_path)
            context.settings.update(storage=new_settings)

            console.print(
                Panel.fit(
                    f"Storage file set to:\n"
                    f"[bold cyan]{storage_path}[/]\n\n"
                    "[dim]Existing secrets were not modified.[/]",
                    title="Storage Updated",
                )
            )

    @app.command(name="encryption")
    def config_encryption(
        mode: Annotated[
            Literal["aesgcm"],
            Option("--mode", "-m", help=ENCRYPTION_MODE_HELP),
        ],
    ):
        """
        Configure the encryption algorithm used by rune.
        """
        context = Context.get()

        if mode == context.settings.encryption.mode:
            display.failed_panel(f"Encryption mode is already set to '[bold]{mode}[/]'.", title="Nothing changed")
            return

        new_settings = EncryptionSettings.from_mode(mode)
        context.settings.update(encryption=new_settings)

        display.success_panel(f"Encryption mode set to '[bold]{mode}[/]'.\n\n" +\
                "[dim]Existing secrets remain encrypted with their original settings.[/]", title="Encryption Updated")

    @app.command(name="show")
    def show_config(
        profile: Annotated[
            Optional[str],
            Argument(
                help=(
                    "Profile name to display.\n"
                    "If omitted, shows the currently active configuration."
                )
            ),
        ] = None,
    ):
        """
        Display the current configuration or a specific profile.
        """
        context = Context.get()

        if not profile:
            settings_file = context.settings_manager.settings_file
            settings = context.settings.to_dict()

            console.print("[bold]Active configuration:[/]")
            console.print(f"Settings file: [cyan]{settings_file}[/]")
            console.print(Panel.fit(Pretty(settings, expand_all=True, indent_guides=True)))
            return

        settings = context.settings_manager.get_profile(profile)

        if not settings:
            display.failed_panel(f"Profile '[bold cyan]{profile}[/]' does not exist.")
            return

        console.print(f"[bold]Configuration for profile '[cyan]{profile}[/]':[/]")
        console.print(Panel.fit(Pretty(settings.to_dict(), expand_all=True, indent_guides=True)))

    @app.command(name="where")
    def whereis(
        interactive: Annotated[
            bool,
            Option(
                "--interactive",
                "-i",
                help="Interactively select a file path to copy to clipboard.",
            ),
        ] = False
    ):
        """
        Show where rune stores its configuration and profile files.
        """
        context = Context.get()
        settings_file = str(context.settings_manager.settings_file.absolute())
        profiles_file = str(context.settings_manager.profiles_file.absolute())

        if not interactive:
            display.panel(
                    f"[bold]Settings file[/]: [cyan]{settings_file}[/]\n"
                    f"[bold]Profiles file[/]: [cyan]{profiles_file}[/]"
            )
            return

        choices = [settings_file, profiles_file]

        display.panel(
            f"[bold cyan][1][/] Settings file: {settings_file}\n"
            f"[bold cyan][2][/] Profiles file: {profiles_file}",
            title="File Locations",
        )

        choice = get_choice_by_idx("Copy file path", choices)
        if choice:
            pyperclip.copy(choice)
            console.print("[green]Path copied to clipboard.[/]")

    app.add_typer(profile_app)




from typing import Dict, List
from rich.console import Console
from rich.panel import Panel
import typer
from rune.commands.getcmd import handle_get_command
from rune.internal.listsecrets import list_secrets
from rich.console import Console
from rich.tree import Tree
import typer

console = Console()

def handle_ls_command(user: str, namespace: str | None, interactive: bool, show: bool, show_deleted: bool):
    result = list_secrets(user)
    secrets = result.value()

    if result.is_failure():
        console.print(Panel.fit(
            f"[bold red]Unable to retrieve secrets:[/] {result.failure_reason()}",
            title="[red]Failed[/]",
        ))
        return

    if not secrets:
        console.print("[yellow]No secrets yet.[/]")
        return
    
    if show_deleted:
        full_names = []
        for secret in secrets:
            fn = secret.full_name
            if secret.deleted:
                parts = fn.split("/")
                parts[-1] = parts[-1] + " (deleted)"
                full_names.append("/".join(parts))
            else:
                full_names.append(fn)
    else:
        full_names = [secret.full_name for secret in secrets if not secret.deleted]

    if not full_names:
        console.print("[yellow]No secrets yet.[/]")
        return

    names_tree = _build_namespace_tree(full_names)

    if namespace:
        for ns in namespace.split("/"):
            if ns != "":
                names_tree = names_tree.get(ns, {})

    if names_tree == {}:
        console.print(f"[yellow]No secrets for namespace {namespace}.[/]")
        return

    compacted_tree = _compact_tree(names_tree)

    root = Tree(f"[bold]{user}/{namespace}/[/]" if namespace else f"[bold]{user}/[/]")
    indexes = _expand_rich_tree(root, compacted_tree)
    console.print(
        Panel.fit(
            root,
            title="[green]✓ Secrets Tree[/]"
        )
    )

    if interactive:
        while True:
            choice = typer.prompt("Select secret to get (q to quit)")
            if choice.lower() == "q":
                break

            try:
                idx = int(choice)
            except:
                continue
            if idx in indexes:
                if namespace:
                    fqn = namespace.removeprefix("/").removesuffix("/").strip() + "/" + indexes[idx]
                else:
                    fqn = indexes[idx]
                console.print(f"Fetching secret [bold]{fqn}[/]")
                handle_get_command(user, _name=fqn, show=show)
                break


def _build_namespace_tree(full_names: List[str]) -> Dict:
    def _build_branch(current: Dict, parts: List[str]) -> Dict:
        if len(parts) == 1:
            if not "__items__" in current:
                current["__items__"] = []
            current["__items__"].append(parts[0])
            return current

        if not parts[0] in current:
            current[parts[0]] = {}

        current[parts[0]] = _build_branch(current.get(parts[0], {}), parts[1:])
        return current

    
    tree = {}
    for path in full_names:
        tree = _build_branch(tree, path.split("/"))

    return tree

def _expand_rich_tree(tree: Tree, node: dict, sorted_item_list: List[str] = [], current_path: str | None = None) -> Dict[int, str]:
    """Recursively expand the namespace tree into a Rich Tree."""
    items = node.get("__items__", [])
    for item in items:
        tree.add(f"[bold cyan][{len(sorted_item_list) + 1}][/] {item.replace("(deleted)", "[bold red][DELETED][/]")}")
        full_name = item if current_path is None else current_path + "/" + item
        sorted_item_list.append(full_name)

    for key, value in node.items():
        if key == "__items__":
            continue
        subtree = tree.add(f"[bold]{key}/[/]")
        full_path = key if not current_path else current_path + "/" + key
        _expand_rich_tree(subtree, value, sorted_item_list, full_path)

    return {idx: item for idx, item in enumerate(sorted_item_list, 1)}

def _compact_tree(node: Dict) -> Dict:
    compacted = {}
    for key, child in node.items():
        if key == "__items__":
            compacted[key] = list(child)
        else:
            compacted[key] = _compact_tree(child)

    result = {}

    for key, child in list(compacted.items()):
        if key == "__items__":
            result.setdefault("__items__", []).extend(child)
            continue

        current_key = key
        current_child = child

        while True:
            if not isinstance(current_child, dict):
                break

            child_items = current_child.get("__items__", [])
            ns_children = [k for k in current_child.keys() if k != "__items__"]

            if child_items:
                break
            if len(ns_children) != 1:
                break

            next_key = ns_children[0]
            next_child = current_child[next_key]

            current_key = f"{current_key}/{next_key}"
            current_child = next_child

        result[current_key] = current_child

    result.setdefault("__items__", [])
    return result




from rich.console import Console
from rich.panel import Panel


console = Console()

def panel(message: str, title: str | None = None) -> None:
    if not title:
        console.print(Panel.fit(message))
    else:
        console.print(Panel.fit(message, title=title))

def success_panel(message: str, title: str = "Success"):
    console.print(Panel.fit(message,title=f"[green]{title}[/]"))

def failed_panel(message: str, title: str = "Failed"):
    console.print(Panel.fit(message,title=f"[red]{title}[/]"))




import typer
from rune.context import Context
from rune.crypto.factory import get_encrypter_by_algorithm
from rune.exception.wrongkey import WrongKeyUsed
from rune.models.result import Failure, Result, Success

def delete_secret(user: str, full_name: str, hard: bool, key: str) -> Result[None]:
    """
    Deletes the encrypted secret via the configured storage manager.

    Returns the reason for failure, if it fails.
    None if is successful.
    """
    storage = Context.get().storage_manager

    secret = storage.retreive_secret(user, full_name)

    if not secret:
        return Failure(f"Secret '{full_name}' does not exist.")

    if not typer.confirm(f"Are you sure you want do delete secret '{full_name}'?"):
        raise typer.Abort()

    if not hard:
        return Success() if storage.delete_secret(user, full_name, hard=False) else Failure(f"Error deleting secret '{full_name}'")

    encrypter = get_encrypter_by_algorithm(secret.algorithm)

    try:
        [encrypter.decrypt(sf, key) for sf in secret.fields.values()]
        return Success() if storage.delete_secret(user, full_name, hard=True) else Failure(f"Error deleting secret '{full_name}'")

    except WrongKeyUsed as e:
        return Failure(e.message)

def restore_secret(user: str, full_name: str) -> Result[None]:
    storage = Context.get().storage_manager

    secret = storage.retreive_secret(user, full_name)

    if not secret:
        return Failure(f"Secret {full_name} does not exist. It might have been hard deleted.")
    
    if not secret.deleted:
        return Failure(f"Secret {full_name} is not deleted.")
    
    return Success() if storage.restore_secret(user, full_name) else Failure("Error restoring secret '{full_name}'")






from rune.context import Context
from rune.models.result import Failure, Result, Success


def move_secret(user: str, original_name: str, new_name: str) -> Result[None]:

    storage = Context.get().storage_manager

    secret = storage.retreive_secret(user, original_name)

    if secret is None:
        return Failure(f"Secret '{original_name}' does not exist")

    if storage.move_secret(user, original_name, new_name):
        return Success()
    else:
        return Failure(f"Unable to move secret '{original_name}'")


